"""
Audio model adapter — load Whisper, Wav2Vec2, HuBERT, and other audio models.

Supports:
- Whisper (OpenAI) — speech recognition encoder
- Wav2Vec2 (Facebook) — self-supervised audio representations
- HuBERT (Facebook) — hidden unit BERT for audio
- WavLM (Microsoft) — speech pre-training with masked prediction
- Other HuggingFace audio models

Usage::

    from model_osmosis.adapters.audio import AudioAdapter

    adapter = AudioAdapter()

    # Load Whisper encoder
    whisper = adapter.load("openai/whisper-base")

    # Load Wav2Vec2
    wav2vec = adapter.load("facebook/wav2vec2-base-960h")

    # Load HuBERT
    hubert = adapter.load("facebook/hubert-base-ls960")
"""

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn as nn
import numpy as np


class AudioModelType(Enum):
    """Types of audio models."""

    WHISPER = "whisper"
    WAV2VEC2 = "wav2vec2"
    HUBERT = "hubert"
    WAVLM = "wavlm"
    GENERIC = "generic"


class AudioExpert(nn.Module):
    """Wrapped audio model as stackable expert.

    Args:
        model: The audio model.
        feature_dim: Output feature dimension.
        name: Human-readable name.
        sample_rate: Expected audio sample rate.
        model_type: Type of audio model.
    """

    def __init__(
        self,
        model: nn.Module,
        feature_dim: int,
        name: str = "audio_expert",
        sample_rate: int = 16000,
        model_type: AudioModelType = AudioModelType.GENERIC,
    ) -> None:
        super().__init__()
        self.model = model
        self.feature_dim = feature_dim
        self.hidden_size = feature_dim  # Alias for compatibility
        self.name = name
        self.sample_rate = sample_rate
        self.model_type = model_type

        # Freeze weights
        for param in self.model.parameters():
            param.requires_grad = False

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Extract features from audio.

        Args:
            x: Audio tensor:
                - Raw waveform (batch, samples)
                - Mel features (batch, features, time) for Whisper
                - Pre-computed embeddings (batch, seq_len, hidden)

        Returns:
            Feature tensor (batch, feature_dim).
        """
        with torch.no_grad():
            if self.model_type == AudioModelType.WHISPER:
                # Whisper uses encoder for feature extraction
                if hasattr(self.model, "encoder"):
                    outputs = self.model.encoder(x)
                else:
                    outputs = self.model(x)
            elif self.model_type in (AudioModelType.WAV2VEC2, AudioModelType.HUBERT, AudioModelType.WAVLM):
                # These take raw waveform as input_values
                if x.dim() == 2:
                    outputs = self.model(input_values=x)
                else:
                    outputs = self.model(x)
            else:
                # Generic: try common input names
                try:
                    outputs = self.model(input_values=x)
                except TypeError:
                    try:
                        outputs = self.model(input_features=x)
                    except TypeError:
                        outputs = self.model(x)

        # Extract hidden states
        if hasattr(outputs, "last_hidden_state"):
            hidden = outputs.last_hidden_state
        elif isinstance(outputs, tuple):
            hidden = outputs[0]
        else:
            hidden = outputs

        # Pool over time dimension
        if hidden.dim() == 3:
            hidden = hidden.mean(dim=1)

        return hidden

    def extra_repr(self) -> str:
        return (
            f"name={self.name}, feature_dim={self.feature_dim}, "
            f"sample_rate={self.sample_rate}, model_type={self.model_type.value}"
        )


class AudioAdapter:
    """Load audio models from HuggingFace.

    Args:
        device: Device to load models to.
        torch_dtype: Data type for model weights.
        verbose: Print loading progress.
    """

    # Model type detection patterns
    MODEL_PATTERNS = {
        AudioModelType.WHISPER: ["whisper"],
        AudioModelType.WAV2VEC2: ["wav2vec2", "wav2vec"],
        AudioModelType.HUBERT: ["hubert"],
        AudioModelType.WAVLM: ["wavlm"],
    }

    # Feature dimensions for common models
    DIM_LOOKUP = {
        "whisper-tiny": 384,
        "whisper-base": 512,
        "whisper-small": 768,
        "whisper-medium": 1024,
        "whisper-large": 1280,
        "wav2vec2-base": 768,
        "wav2vec2-large": 1024,
        "hubert-base": 768,
        "hubert-large": 1024,
        "hubert-xlarge": 1280,
        "wavlm-base": 768,
        "wavlm-large": 1024,
    }

    def __init__(
        self,
        device: str = "cpu",
        torch_dtype: Optional[torch.dtype] = None,
        verbose: bool = True,
    ) -> None:
        self.device = device
        self.torch_dtype = torch_dtype
        self.verbose = verbose

    def load(
        self,
        model_name: str,
        name: Optional[str] = None,
        **kwargs,
    ) -> AudioExpert:
        """Load an audio model.

        Args:
            model_name: HuggingFace model ID.
            name: Human-readable name.

        Returns:
            AudioExpert ready for stacking.
        """
        try:
            from transformers import AutoConfig
        except ImportError:
            raise ImportError(
                "Audio adapter requires `transformers`. "
                "Install with: pip install transformers"
            )

        if self.verbose:
            print(f"[Audio] Loading '{model_name}'...")

        # Detect model type
        model_type = self._detect_type(model_name)

        # Load config
        config = AutoConfig.from_pretrained(model_name)

        # Get feature dimension
        feature_dim = self._get_feature_dim(config, model_name)

        # Get sample rate
        sample_rate = getattr(config, "sample_rate", 16000)

        # Load model
        model = self._load_model(model_name, model_type, config, kwargs)

        model = model.to(self.device)
        if self.torch_dtype:
            model = model.to(self.torch_dtype)
        model.eval()

        if self.verbose:
            total_params = sum(p.numel() for p in model.parameters())
            print(
                f"[Audio] Loaded '{name or model_name.split('/')[-1]}': "
                f"{total_params:,} params, feature_dim={feature_dim}, "
                f"sample_rate={sample_rate}"
            )

        return AudioExpert(
            model=model,
            feature_dim=feature_dim,
            name=name or model_name.split("/")[-1],
            sample_rate=sample_rate,
            model_type=model_type,
        )

    def _detect_type(self, model_name: str) -> AudioModelType:
        """Detect audio model type from name."""
        name_lower = model_name.lower()
        for model_type, patterns in self.MODEL_PATTERNS.items():
            for pattern in patterns:
                if pattern in name_lower:
                    return model_type
        return AudioModelType.GENERIC

    def _get_feature_dim(self, config: Any, model_name: str) -> int:
        """Get feature dimension from config or lookup."""
        # Try config attributes
        for attr in ["hidden_size", "d_model", "encoder_hidden_size"]:
            dim = getattr(config, attr, None)
            if dim is not None:
                return dim

        # Try lookup table
        name_lower = model_name.lower()
        for key, dim in self.DIM_LOOKUP.items():
            if key in name_lower:
                return dim

        # Default
        return 768

    def _load_model(
        self,
        model_name: str,
        model_type: AudioModelType,
        config: Any,
        kwargs: Dict,
    ) -> nn.Module:
        """Load the appropriate model class."""
        from transformers import AutoModel

        if model_type == AudioModelType.WHISPER:
            # For Whisper, we want the encoder model
            from transformers import WhisperModel
            model = WhisperModel.from_pretrained(model_name, **kwargs)
        else:
            model = AutoModel.from_pretrained(model_name, **kwargs)

        return model

    def load_multiple(
        self,
        model_names: List[str],
        names: Optional[List[str]] = None,
    ) -> List[AudioExpert]:
        """Load multiple audio models.

        Args:
            model_names: List of HuggingFace model IDs.
            names: Optional list of names.

        Returns:
            List of AudioExpert modules.
        """
        names = names or [None] * len(model_names)
        experts = []
        for model_name, name in zip(model_names, names):
            expert = self.load(model_name, name=name)
            experts.append(expert)
        return experts

    def get_feature_dim(self, model_name: str) -> int:
        """Get feature dimension without loading full model.

        Args:
            model_name: HuggingFace model ID.

        Returns:
            Feature dimension.
        """
        try:
            from transformers import AutoConfig
            config = AutoConfig.from_pretrained(model_name)
            return self._get_feature_dim(config, model_name)
        except Exception:
            # Try lookup
            name_lower = model_name.lower()
            for key, dim in self.DIM_LOOKUP.items():
                if key in name_lower:
                    return dim
            return 768



"""
Tests for multi-model merging.
"""

import numpy as np
import pytest
import torch
import torch.nn as nn

from model_osmosis.merging.strategies import (
    WeightedAverageStrategy,
    SLERPStrategy,
    TIESStrategy,
    get_strategy,
)
from model_osmosis.merging.streaming import StreamingParameterLoader
from model_osmosis.merging.evaluator import MergeEvaluator
from model_osmosis.merging.merger import MultiModelMerger
from model_osmosis import OsmosisPipeline, OsmosisConfig


# ---------------------------------------------------------------------------
# Fixtures
# ---------------------------------------------------------------------------


def _make_model(seed: int = 0):
    """Create a simple model with deterministic weights."""
    torch.manual_seed(seed)
    model = nn.Sequential(
        nn.Linear(6, 16),
        nn.ReLU(),
        nn.Linear(16, 8),
        nn.ReLU(),
        nn.Linear(8, 3),
    )
    for p in model.parameters():
        if p.dim() > 1:
            nn.init.xavier_normal_(p)
    model.eval()
    return model


@pytest.fixture
def three_models():
    """Three models with different weights."""
    return [_make_model(seed=i) for i in range(3)]


@pytest.fixture
def sample_data():
    return np.random.rand(50, 6).astype(np.float32)


# ---------------------------------------------------------------------------
# Strategy Tests
# ---------------------------------------------------------------------------


class TestWeightedAverage:
    def test_uniform_average(self):
        strat = WeightedAverageStrategy()
        a = np.array([1.0, 2.0, 3.0])
        b = np.array([5.0, 6.0, 7.0])
        result = strat.merge_tensors([a, b])
        np.testing.assert_allclose(result, [3.0, 4.0, 5.0])

    def test_weighted_average(self):
        strat = WeightedAverageStrategy()
        a = np.array([0.0, 0.0])
        b = np.array([10.0, 10.0])
        result = strat.merge_tensors([a, b], weights=[0.8, 0.2])
        np.testing.assert_allclose(result, [2.0, 2.0])

    def test_three_tensors(self):
        strat = WeightedAverageStrategy()
        a = np.ones(5) * 1
        b = np.ones(5) * 2
        c = np.ones(5) * 3
        result = strat.merge_tensors([a, b, c])
        np.testing.assert_allclose(result, [2.0] * 5)


class TestSLERP:
    def test_two_tensors(self):
        strat = SLERPStrategy()
        a = np.array([1.0, 0.0, 0.0])
        b = np.array([0.0, 1.0, 0.0])
        result = strat.merge_tensors([a, b], weights=[0.5, 0.5])
        # SLERP at t=0.5 should be on the arc between a and b
        assert result.shape == (3,)
        # Both components should be positive and similar
        assert result[0] > 0 and result[1] > 0

    def test_identical_tensors(self):
        strat = SLERPStrategy()
        a = np.array([1.0, 2.0, 3.0])
        result = strat.merge_tensors([a, a])
        np.testing.assert_allclose(result, a, atol=1e-5)

    def test_three_tensors(self):
        strat = SLERPStrategy()
        tensors = [np.random.rand(10) for _ in range(3)]
        result = strat.merge_tensors(tensors)
        assert result.shape == (10,)


class TestTIES:
    def test_basic_merge(self):
        strat = TIESStrategy(density=0.5)
        a = np.array([1.0, -2.0, 3.0, -4.0])
        b = np.array([1.5, -1.5, 2.5, -3.5])
        c = np.array([0.5, -2.5, 3.5, -4.5])
        result = strat.merge_tensors([a, b, c])
        assert result.shape == (4,)

    def test_density_parameter(self):
        strat_high = TIESStrategy(density=0.9)
        strat_low = TIESStrategy(density=0.1)
        tensors = [np.random.rand(20) for _ in range(3)]
        r_high = strat_high.merge_tensors(tensors)
        r_low = strat_low.merge_tensors(tensors)
        # Both should produce valid outputs
        assert r_high.shape == (20,)
        assert r_low.shape == (20,)


class TestStrategyRegistry:
    def test_get_valid_strategy(self):
        s = get_strategy("weighted_average")
        assert isinstance(s, WeightedAverageStrategy)

    def test_get_invalid_raises(self):
        with pytest.raises(ValueError):
            get_strategy("nonexistent")


# ---------------------------------------------------------------------------
# Streaming Tests
# ---------------------------------------------------------------------------


class TestStreamingParameterLoader:
    def test_stream_from_module(self):
        model = _make_model()
        loader = StreamingParameterLoader()
        params = list(loader.stream_parameters(model))

        assert len(params) > 0
        for name, arr in params:
            assert isinstance(name, str)
            assert isinstance(arr, np.ndarray)

    def test_stream_from_dict(self):
        model = _make_model()
        state_dict = {k: v.cpu() for k, v in model.state_dict().items()}
        loader = StreamingParameterLoader()
        params = list(loader.stream_parameters(state_dict))
        assert len(params) > 0

    def test_get_param_names(self):
        model = _make_model()
        loader = StreamingParameterLoader()
        names = loader.get_all_param_names(model)
        assert len(names) > 0
        assert all(isinstance(n, str) for n in names)


# ---------------------------------------------------------------------------
# Evaluator Tests
# ---------------------------------------------------------------------------


class TestMergeEvaluator:
    def test_find_best_weights(self, three_models, sample_data):
        evaluator = MergeEvaluator()
        strat = WeightedAverageStrategy()
        weights, score = evaluator.find_best_weights(
            three_models, sample_data, strat, n_trials=5
        )
        assert len(weights) == 3
        assert all(w >= 0 for w in weights)
        assert abs(sum(weights) - 1.0) < 1e-5

    def test_single_model_returns_default(self):
        evaluator = MergeEvaluator()
        model = _make_model()
        strat = WeightedAverageStrategy()
        weights, score = evaluator.find_best_weights(
            [model], np.random.rand(10, 6).astype(np.float32), strat
        )
        assert weights == [1.0]


# ---------------------------------------------------------------------------
# MultiModelMerger Tests
# ---------------------------------------------------------------------------


class TestMultiModelMerger:
    def test_merge_weighted_average(self, three_models):
        merger = MultiModelMerger(strategy="weighted_average", verbose=False)
        result = merger.merge(three_models)

        # Should produce a valid model
        x = torch.rand(5, 6)
        with torch.no_grad():
            out = result(x)
        assert out.shape == (5, 3)

    def test_merge_slerp(self, three_models):
        merger = MultiModelMerger(strategy="slerp", verbose=False)
        result = merger.merge(three_models)

        x = torch.rand(5, 6)
        with torch.no_grad():
            out = result(x)
        assert out.shape == (5, 3)

    def test_merge_ties(self, three_models):
        merger = MultiModelMerger(strategy="ties", verbose=False)
        result = merger.merge(three_models)

        x = torch.rand(5, 6)
        with torch.no_grad():
            out = result(x)
        assert out.shape == (5, 3)

    def test_merge_auto(self, three_models, sample_data):
        merger = MultiModelMerger(strategy="auto", verbose=False)
        result = merger.merge(three_models, eval_data=sample_data)

        x = torch.rand(5, 6)
        with torch.no_grad():
            out = result(x)
        assert out.shape == (5, 3)

    def test_merge_custom_weights(self, three_models):
        merger = MultiModelMerger(
            strategy="weighted_average",
            model_weights=[0.6, 0.3, 0.1],
            verbose=False,
        )
        result = merger.merge(three_models)

        x = torch.rand(5, 6)
        with torch.no_grad():
            out = result(x)
        assert out.shape == (5, 3)

    def test_merged_differs_from_originals(self, three_models):
        """Merged model should produce different outputs than any single model."""
        merger = MultiModelMerger(strategy="weighted_average", verbose=False)
        result = merger.merge(three_models)

        x = torch.rand(10, 6)
        with torch.no_grad():
            merged_out = result(x).numpy()
            original_outs = [m(x).numpy() for m in three_models]

        # Merged should not be identical to any single model
        for orig in original_outs:
            assert not np.allclose(merged_out, orig, atol=1e-6)


# ---------------------------------------------------------------------------
# Pipeline.merge() Tests
# ---------------------------------------------------------------------------


class TestPipelineMerge:
    def test_pipeline_merge(self, three_models):
        pipeline = OsmosisPipeline(OsmosisConfig(verbose=False))
        result = pipeline.merge(
            three_models,
            strategy="weighted_average",
        )

        x = torch.rand(5, 6)
        with torch.no_grad():
            out = result(x)
        assert out.shape == (5, 3)

    def test_pipeline_merge_auto(self, three_models, sample_data):
        pipeline = OsmosisPipeline(OsmosisConfig(verbose=False))
        result = pipeline.merge(
            three_models,
            strategy="auto",
            eval_data=sample_data,
        )

        x = torch.rand(5, 6)
        with torch.no_grad():
            out = result(x)
        assert out.shape == (5, 3)



"""Unified deterministic self-optimization state for local-only learning."""

from __future__ import annotations

import json
import os
import shutil
import time
from pathlib import Path
from typing import Any, Dict, Optional


OPTIMIZER_MODES = {"auto", "off"}
STATE_SCHEMA_VERSION = 1
DEFAULT_SUBSYSTEMS = (
    "strategy",
    "compression",
    "protection",
    "cross_merge_classic",
    "cross_merge_llm",
    "expand",
)
CANARY_STATUSES = {"shadow", "canary", "promoted", "rollback"}


def _utc_now() -> str:
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())


def optimizer_home() -> Path:
    env = str(os.environ.get("OSMOSIS_OPTIMIZER_HOME", "")).strip()
    if env:
        return Path(env)
    return Path(".model_osmosis") / "self_optimize"


def state_file_path() -> Path:
    return optimizer_home() / "state.json"


def _subsystem_default() -> Dict[str, Any]:
    return {
        "canary_status": "shadow",
        "promoted_profile_id": "",
        "active_profile_id": "",
        "rollback_reason": "",
        "pack_version": "",
        "selector_version": "",
        "last_selection_latency_ms": 0.0,
        "runs": 0,
        "shadow_runs": 0,
        "canary_attempts": 0,
        "canary_passes": 0,
        "candidates": {},
        "scores": {},
    }


def _default_state() -> Dict[str, Any]:
    return {
        "schema_version": STATE_SCHEMA_VERSION,
        "updated_at": "",
        "rollback_marker": {},
        "subsystems": {name: _subsystem_default() for name in DEFAULT_SUBSYSTEMS},
    }


def _normalize_state(state: Dict[str, Any]) -> Dict[str, Any]:
    normalized = _default_state()
    normalized["schema_version"] = int(state.get("schema_version", STATE_SCHEMA_VERSION) or STATE_SCHEMA_VERSION)
    normalized["updated_at"] = str(state.get("updated_at", ""))
    rollback = state.get("rollback_marker", {})
    if isinstance(rollback, dict):
        normalized["rollback_marker"] = {
            "subsystem": str(rollback.get("subsystem", "")),
            "profile_id": str(rollback.get("profile_id", "")),
            "reason": str(rollback.get("reason", "")),
            "at": str(rollback.get("at", "")),
        }

    subsystems_in = state.get("subsystems", {})
    if isinstance(subsystems_in, dict):
        for name, value in subsystems_in.items():
            if not isinstance(name, str) or not isinstance(value, dict):
                continue
            entry = _subsystem_default()
            entry["canary_status"] = (
                str(value.get("canary_status", "shadow"))
                if str(value.get("canary_status", "shadow")) in CANARY_STATUSES
                else "shadow"
            )
            entry["promoted_profile_id"] = str(value.get("promoted_profile_id", ""))
            entry["active_profile_id"] = str(value.get("active_profile_id", ""))
            entry["rollback_reason"] = str(value.get("rollback_reason", ""))
            entry["pack_version"] = str(value.get("pack_version", ""))
            entry["selector_version"] = str(value.get("selector_version", ""))
            entry["last_selection_latency_ms"] = float(value.get("last_selection_latency_ms", 0.0) or 0.0)
            entry["runs"] = int(value.get("runs", 0) or 0)
            entry["shadow_runs"] = int(value.get("shadow_runs", 0) or 0)
            entry["canary_attempts"] = int(value.get("canary_attempts", 0) or 0)
            entry["canary_passes"] = int(value.get("canary_passes", 0) or 0)
            candidates = value.get("candidates", {})
            if isinstance(candidates, dict):
                entry["candidates"] = {
                    str(k): int(v or 0)
                    for k, v in sorted(candidates.items(), key=lambda item: str(item[0]))
                }
            scores = value.get("scores", {})
            if isinstance(scores, dict):
                entry["scores"] = {
                    str(k): float(v or 0.0)
                    for k, v in sorted(scores.items(), key=lambda item: str(item[0]))
                }
            normalized["subsystems"][str(name)] = entry
    return normalized


def load_state() -> Dict[str, Any]:
    path = state_file_path()
    if not path.exists():
        return _default_state()
    try:
        parsed = json.loads(path.read_text(encoding="utf-8"))
    except Exception:
        return _default_state()
    if not isinstance(parsed, dict):
        return _default_state()
    return _normalize_state(parsed)


def save_state(state: Dict[str, Any]) -> None:
    normalized = _normalize_state(state)
    normalized["updated_at"] = _utc_now()
    path = state_file_path()
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(normalized, indent=2, sort_keys=True), encoding="utf-8")


def resolve_optimizer_mode(requested: Optional[str]) -> str:
    mode = str(requested or "auto").strip().lower()
    env_mode = str(os.environ.get("OSMOSIS_OPTIMIZER", "")).strip().lower()
    if env_mode in OPTIMIZER_MODES:
        mode = env_mode
    if mode not in OPTIMIZER_MODES:
        raise ValueError(
            f"Unsupported optimizer mode '{mode}'. Expected one of: {sorted(OPTIMIZER_MODES)}"
        )
    return mode


def resolve_subsystem_mode(
    requested: Optional[str],
    *,
    subsystem_env_var: Optional[str] = None,
) -> str:
    global_mode = resolve_optimizer_mode("auto")
    if global_mode == "off":
        return "off"

    mode = str(requested or "auto").strip().lower()
    if subsystem_env_var:
        env_mode = str(os.environ.get(subsystem_env_var, "")).strip().lower()
        if env_mode in OPTIMIZER_MODES:
            mode = env_mode
    if mode not in OPTIMIZER_MODES:
        raise ValueError(
            f"Unsupported mode '{mode}'. Expected one of: {sorted(OPTIMIZER_MODES)}"
        )
    return mode


def record_subsystem_selection(
    *,
    subsystem: str,
    profile_id: str,
    pack_version: str,
    selector_version: str,
    selection_latency_ms: float,
    quality_ok: bool,
    fallback_reason: str = "",
    quality_score: Optional[float] = None,
) -> Dict[str, Any]:
    name = str(subsystem or "").strip().lower() or "unknown"
    pid = str(profile_id or "").strip() or "unknown"
    state = load_state()
    subsystems = state.setdefault("subsystems", {})
    entry = subsystems.setdefault(name, _subsystem_default())

    entry["runs"] = int(entry.get("runs", 0) or 0) + 1
    entry["active_profile_id"] = pid
    entry["pack_version"] = str(pack_version or "")
    entry["selector_version"] = str(selector_version or "")
    entry["last_selection_latency_ms"] = round(float(selection_latency_ms or 0.0), 4)

    candidates = entry.setdefault("candidates", {})
    candidates[pid] = int(candidates.get(pid, 0) or 0) + 1

    if quality_score is not None:
        scores = entry.setdefault("scores", {})
        prev = float(scores.get(pid, 0.0) or 0.0)
        scores[pid] = round(0.8 * prev + 0.2 * float(quality_score), 6)

    if not quality_ok:
        entry["canary_status"] = "rollback"
        entry["rollback_reason"] = str(fallback_reason or "quality_regression")
        entry["canary_passes"] = 0
        state["rollback_marker"] = {
            "subsystem": name,
            "profile_id": pid,
            "reason": entry["rollback_reason"],
            "at": _utc_now(),
        }
    else:
        current = str(entry.get("canary_status", "shadow"))
        promoted = str(entry.get("promoted_profile_id", ""))
        if promoted == pid:
            entry["canary_status"] = "promoted"
            entry["rollback_reason"] = ""
            entry["canary_passes"] = int(entry.get("canary_passes", 0) or 0) + 1
        else:
            passes = int(entry.get("canary_passes", 0) or 0) + 1
            entry["canary_passes"] = passes
            if current in {"shadow", "rollback"}:
                entry["canary_status"] = "canary"
            elif current == "canary" and passes >= 2:
                entry["canary_status"] = "promoted"
                entry["promoted_profile_id"] = pid
                entry["rollback_reason"] = ""

    save_state(state)
    return get_subsystem_metadata(name)


def get_subsystem_metadata(subsystem: str) -> Dict[str, Any]:
    name = str(subsystem or "").strip().lower() or "unknown"
    state = load_state()
    entry = dict(state.get("subsystems", {}).get(name, _subsystem_default()))
    return {
        "canary_status": str(entry.get("canary_status", "shadow")),
        "promoted_profile_id": str(entry.get("promoted_profile_id", "")),
        "active_profile_id": str(entry.get("active_profile_id", "")),
        "rollback_reason": str(entry.get("rollback_reason", "")),
        "pack_version": str(entry.get("pack_version", "")),
        "selector_version": str(entry.get("selector_version", "")),
        "selection_latency_ms": round(float(entry.get("last_selection_latency_ms", 0.0) or 0.0), 4),
        "runs": int(entry.get("runs", 0) or 0),
        "shadow_runs": int(entry.get("shadow_runs", 0) or 0),
        "canary_attempts": int(entry.get("canary_attempts", 0) or 0),
        "canary_passes": int(entry.get("canary_passes", 0) or 0),
    }


def get_optimizer_status() -> Dict[str, Any]:
    state = load_state()
    subsystems_raw = state.get("subsystems", {})
    status_subsystems: Dict[str, Any] = {}
    if isinstance(subsystems_raw, dict):
        for name, _ in sorted(subsystems_raw.items(), key=lambda item: str(item[0])):
            status_subsystems[str(name)] = get_subsystem_metadata(str(name))
    return {
        "schema_version": STATE_SCHEMA_VERSION,
        "optimizer_mode": resolve_optimizer_mode("auto"),
        "state_path": str(state_file_path()),
        "updated_at": str(state.get("updated_at", "")),
        "rollback_marker": dict(state.get("rollback_marker", {}) or {}),
        "subsystems": status_subsystems,
    }


def reset_optimizer_state() -> Dict[str, Any]:
    home = optimizer_home()
    if home.exists():
        shutil.rmtree(home)
    return get_optimizer_status()
